@@ -23,8 +23,8 @@
 #include <wctype.h>
 #include <fnmatch.h>
 #include <sys/mman.h>
+#include <libgen.h>
 
-#include "basenames.h"
 #include "log.h"
 #include "logrotate.h"
 
@@ -33,6 +33,7 @@
 #endif
 
 #define REALLOC_STEP    10
+#define GLOB_STR_REALLOC_STEP	0x100
 
 #if defined(SunOS)
 #include <limits.h>
@@ -45,7 +46,7 @@
 #include "asprintf.c"
 #endif
 
-#if !defined(asprintf) && !defined(_FORTIFY_SOURCE)
+#if !defined(HAVE_ASPRINTF) && !defined(_FORTIFY_SOURCE)
 #include <stdarg.h>
 
 int asprintf(char **string_ptr, const char *format, ...)
@@ -78,7 +79,7 @@ int asprintf(char **string_ptr, const char *format, ...)
 
 #endif
 
-#if !defined(strndup)
+#if !defined(HAVE_STRNDUP)
 char *strndup(const char *s, size_t n)
 {
        size_t nAvail;
@@ -101,6 +102,21 @@ char *strndup(const char *s, size_t n)
 }
 #endif
 
+/* list of compression commands and the corresponding file extensions */
+struct compress_cmd_item {
+    const char *cmd;
+    const char *ext;
+};
+static const struct compress_cmd_item compress_cmd_list[] = {
+    {"gzip", ".gz"},
+    {"bzip2", ".bz2"},
+    {"xz", ".xz"},
+    {"compress", ".Z"},
+    {"zip", "zip"},
+};
+static const int compress_cmd_list_size = sizeof(compress_cmd_list)
+    / sizeof(compress_cmd_list[0]);
+
 enum {
 	STATE_DEFAULT = 2,
 	STATE_SKIP_LINE = 4,
@@ -110,7 +126,7 @@ enum {
 	STATE_ERROR = 64,
 };
 
-static char *defTabooExts[] = { ".rpmsave", ".rpmorig", "~", ",v",
+static const char *defTabooExts[] = { ".rpmsave", ".rpmorig", "~", ",v",
     ".disabled", ".dpkg-old", ".dpkg-dist", ".dpkg-new", ".cfsaved",
     ".ucf-old", ".ucf-dist", ".ucf-new",
     ".rpmnew", ".swp", ".cfsaved", ".rhn-cfg-tmp-*"
@@ -118,8 +134,8 @@ static char *defTabooExts[] = { ".rpmsave", ".rpmorig", "~", ",v",
 static int defTabooCount = sizeof(defTabooExts) / sizeof(char *);
 
 /* I shouldn't use globals here :-( */
-static char **tabooExts = NULL;
-int tabooCount = 0;
+static char **tabooPatterns = NULL;
+static int tabooCount = 0;
 static int glob_errno = 0;
 
 static int readConfigFile(const char *configFile, struct logInfo *defConfig);
@@ -127,34 +143,38 @@ static int globerr(const char *pathname, int theerr);
 
 static char *isolateLine(char **strt, char **buf, size_t length) {
 	char *endtag, *start, *tmp;
+	const char *max = *buf + length;
+	char *key;
+
 	start = *strt;
 	endtag = start;
-	while (endtag - *buf < length && *endtag != '\n') {
+	while (endtag < max && *endtag != '\n') {
 		endtag++;}
-	if (endtag - *buf > length)
+	if (max < endtag)
 		return NULL;
 	tmp = endtag - 1;
 	while (isspace((unsigned char)*endtag))
 		endtag--;
-	char *key = strndup(start, endtag - start + 1);
+	key = strndup(start, endtag - start + 1);
 	*strt = tmp;
 	return key;
 }
 
-static char *isolateValue(const char *fileName, int lineNum, char *key,
+static char *isolateValue(const char *fileName, int lineNum, const char *key,
 			char **startPtr, char **buf, size_t length)
 {
     char *chptr = *startPtr;
+    const char *max = *startPtr + length;
 
-    while (chptr - *buf < length && isblank((unsigned char)*chptr))
+    while (chptr < max && isblank((unsigned char)*chptr))
 	chptr++;
-    if (chptr - *buf < length && *chptr == '=') {
+    if (chptr < max && *chptr == '=') {
 	chptr++;
-	while ( chptr - *buf < length && isblank((unsigned char)*chptr))
+	while ( chptr < max && isblank((unsigned char)*chptr))
 	    chptr++;
     }
 
-    if (chptr - *buf < length && *chptr == '\n') {
+    if (chptr < max && *chptr == '\n') {
 		message(MESS_ERROR, "%s:%d argument expected after %s\n",
 			fileName, lineNum, key);
 		return NULL;
@@ -166,24 +186,26 @@ static char *isolateValue(const char *fileName, int lineNum, char *key,
 
 static char *isolateWord(char **strt, char **buf, size_t length) {
 	char *endtag, *start;
+	const char *max = *buf + length;
+	char *key;
 	start = *strt;
-	while (start - *buf < length && isblank((unsigned char)*start))
+	while (start < max && isblank((unsigned char)*start))
 		start++;
 	endtag = start;
-	while (endtag - *buf < length && isalpha((unsigned char)*endtag)) {
+	while (endtag < max && isalpha((unsigned char)*endtag)) {
 		endtag++;}
-	if (endtag - *buf > length)
+	if (max < endtag)
 		return NULL;
-	char *key = strndup(start, endtag - start);
+	key = strndup(start, endtag - start);
 	*strt = endtag;
 	return key;
 }
 
-static char *readPath(const char *configFile, int lineNum, char *key,
+static char *readPath(const char *configFile, int lineNum, const char *key,
 		      char **startPtr, char **buf, size_t length)
 {
     char *chptr;
-    char *start = *startPtr;
+    char *start;
     char *path;
 
     wchar_t pwc;
@@ -224,13 +246,18 @@ static int readModeUidGid(const char *configFile, int lineNum, char *key,
 							const char *directive, mode_t *mode, uid_t *uid,
 							gid_t *gid) {
 	char u[200], g[200];
-	int m;
+	unsigned int m;
 	char tmp;
 	int rc;
 	struct group *group;
 	struct passwd *pw = NULL;
 
-	rc = sscanf(key, "%o %199s %199s%c", &m, u, g, &tmp);
+	if (!strcmp("su", directive))
+	    /* do not read <mode> for the 'su' directive */
+	    rc = 0;
+	else
+	    rc = sscanf(key, "%o %199s %199s%c", &m, u, g, &tmp);
+
 	/* We support 'key <owner> <group> notation now */
 	if (rc == 0) {
 		rc = sscanf(key, "%199s %199s%c", u, g, &tmp);
@@ -275,7 +302,7 @@ static int readModeUidGid(const char *configFile, int lineNum, char *key,
 	return 0;
 }
 
-static char *readAddress(const char *configFile, int lineNum, char *key,
+static char *readAddress(const char *configFile, int lineNum, const char *key,
 			 char **startPtr, char **buf, size_t length)
 {
     char *endtag, *chptr;
@@ -313,13 +340,17 @@ static int do_mkdir(const char *path, mode_t mode, uid_t uid, gid_t gid) {
 				path, strerror(errno));
 			return -1;
 		}
-		if ((uid != sb.st_uid || gid != sb.st_gid) && 
-			chown(path, uid, gid)) {
+		if (chown(path, uid, gid) != 0) {
 			message(MESS_ERROR, "error setting owner of %s to uid %d and gid %d: %s\n",
 				path, uid, gid, strerror(errno));
 			return -1;
 		}
-    }
+		if (chmod(path, mode) != 0) {
+			message(MESS_ERROR, "error setting permissions of %s to 0%o: %s\n",
+				path, mode, strerror(errno));
+			return -1;
+		}
+	}
 	else if (!S_ISDIR(sb.st_mode)) {
 		message(MESS_ERROR, "path %s already exists, but it is not a directory\n",
 			path);
@@ -338,7 +369,7 @@ static int mkpath(const char *path, mode_t mode, uid_t uid, gid_t gid) {
 
 	rv = 0;
 	pp = copypath;
-	while (rv == 0 && (sp = strchr(pp, '/')) != 0) {
+	while (rv == 0 && (sp = strchr(pp, '/')) != NULL) {
 		if (sp != pp) {
 			*sp = '\0';
 			rv = do_mkdir(copypath, mode, uid, gid);
@@ -356,7 +387,6 @@ static int mkpath(const char *path, mode_t mode, uid_t uid, gid_t gid) {
 static int checkFile(const char *fname)
 {
 	int i;
-	char *pattern;
 
 	/* Check if fname is '.' or '..'; if so, return false */
 	if (fname[0] == '.' && (!fname[1] || (fname[1] == '.' && !fname[2])))
@@ -364,18 +394,14 @@ static int checkFile(const char *fname)
 
 	/* Check if fname is ending in a taboo-extension; if so, return false */
 	for (i = 0; i < tabooCount; i++) {
-		if (asprintf(&pattern, "*%s", tabooExts[i]) < 0) {
-			message(MESS_FATAL, "failed to allocate taboo pattern memory\n");
-		}
-		if (!fnmatch(pattern, fname, 0))
+		const char *pattern = tabooPatterns[i];
+		if (!fnmatch(pattern, fname, FNM_PERIOD))
 		{
-			free(pattern);
-			message(MESS_DEBUG, "Ignoring %s, because of %s ending\n",
-					fname, tabooExts[i]);
+			message(MESS_DEBUG, "Ignoring %s, because of %s pattern match\n",
+					fname, pattern);
 			return 0;
 		}
 	}
-	free(pattern);
 	/* All checks have been passed; return true */
 	return 1;
 }
@@ -401,10 +427,12 @@ static void copyLogInfo(struct logInfo *to, struct logInfo *from)
     if (from->oldDir)
 	to->oldDir = strdup(from->oldDir);
     to->criterium = from->criterium;
+    to->weekday = from->weekday;
     to->threshhold = from->threshhold;
     to->minsize = from->minsize;
-	to->maxsize = from->maxsize;
+    to->maxsize = from->maxsize;
     to->rotateCount = from->rotateCount;
+    to->rotateMinAge = from->rotateMinAge;
     to->rotateAge = from->rotateAge;
     to->logStart = from->logStart;
     if (from->pre)
@@ -500,8 +528,8 @@ static void freeTailLogs(int num)
 static int readConfigPath(const char *path, struct logInfo *defConfig)
 {
     struct stat sb;
-    int here, oldnumlogs, result = 1;
-	struct logInfo defConfigBackup;
+    int here, result = 0;
+    struct logInfo defConfigBackup;
 
     if (stat(path, &sb)) {
 	message(MESS_ERROR, "cannot stat %s: %s\n", path, strerror(errno));
@@ -578,17 +606,14 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
 
 	for (i = 0; i < files_count; ++i) {
 	    assert(namelist[i] != NULL);
-	    oldnumlogs = numLogs;
 	    copyLogInfo(&defConfigBackup, defConfig);
 	    if (readConfigFile(namelist[i], defConfig)) {
 		message(MESS_ERROR, "found error in file %s, skipping\n", namelist[i]);
-		freeTailLogs(numLogs - oldnumlogs);
 		freeLogInfo(defConfig);
 		copyLogInfo(defConfig, &defConfigBackup);
 		freeLogInfo(&defConfigBackup);
+		result = 1;
 		continue;
-	    } else {
-		result = 0;
 	    }
 	    freeLogInfo(&defConfigBackup);
 	}
@@ -599,14 +624,11 @@ static int readConfigPath(const char *path, struct logInfo *defConfig)
 	close(here);
 	free_2d_array(namelist, files_count);
     } else {
-    	oldnumlogs = numLogs;
 	copyLogInfo(&defConfigBackup, defConfig);
 	if (readConfigFile(path, defConfig)) {
-	    freeTailLogs(numLogs - oldnumlogs);
 	    freeLogInfo(defConfig);
 	    copyLogInfo(defConfig, &defConfigBackup);
-	} else {
-	    result = 0;
+	    result = 1;
 	}
 	freeLogInfo(&defConfigBackup);
     }
@@ -628,6 +650,7 @@ int readAllConfigPaths(const char **paths)
 		.minsize = 0,
 		.maxsize = 0,
 		.rotateCount = 0,
+		.rotateMinAge = 0,
 		.rotateAge = 0,
 		.logStart = -1,
 		.pre = NULL,
@@ -637,6 +660,7 @@ int readAllConfigPaths(const char **paths)
 		.preremove = NULL,
 		.logAddress = NULL,
 		.extension = NULL,
+		.addextension = NULL,
 		.compress_prog = NULL,
 		.uncompress_prog = NULL,
 		.compress_ext = NULL,
@@ -655,31 +679,110 @@ int readAllConfigPaths(const char **paths)
 		.compress_options_count = 0
     };
 
-    tabooExts = malloc(sizeof(*tabooExts) * defTabooCount);
+    tabooPatterns = malloc(sizeof(*tabooPatterns) * defTabooCount);
     for (i = 0; i < defTabooCount; i++) {
-	if ((tabooExts[i] = (char *) malloc(strlen(defTabooExts[i]) + 1))) {
-	    strcpy(tabooExts[i], defTabooExts[i]);
+	int bytes;
+	char *pattern = NULL;
+
+	/* generate a pattern by concatenating star (wildcard) to the
+	 * suffix literal
+	 */
+	bytes = asprintf(&pattern, "*%s", defTabooExts[i]);
+	if (bytes != -1) {
+	    tabooPatterns[i] = pattern;
 	    tabooCount++;
 	} else {
-	    free_2d_array(tabooExts, tabooCount);
+	    free_2d_array(tabooPatterns, tabooCount);
 	    message(MESS_ERROR, "cannot malloc: %s\n", strerror(errno));
 	    return 1;
 	}
     }
 
     for (file = paths; *file; file++) {
-	if (readConfigPath(*file, &defConfig)) {
+	if (readConfigPath(*file, &defConfig))
 	    result = 1;
-	    break;
-	}
     }
-    free_2d_array(tabooExts, tabooCount);
+    free_2d_array(tabooPatterns, tabooCount);
     freeLogInfo(&defConfig);
     return result;
 }
 
+static char* parseGlobString(const char *configFile, int lineNum,
+			     const char *buf, off_t length, char **ppos)
+{
+    /* output buffer */
+    char *globString = NULL;
+    size_t globStringPos = 0;
+    size_t globStringAlloc = 0;
+    enum {
+	PGS_INIT,	/* picking blanks, looking for '#' */
+	PGS_DATA,	/* picking data, looking for end of line */
+	PGS_COMMENT	/* skipping comment, looking for end of line */
+    } state = PGS_INIT;
+
+    /* move the cursor at caller's side while going through the input */
+    for (; (*ppos - buf < length) && **ppos; (*ppos)++) {
+	/* state transition (see above) */
+	switch (state) {
+	    case PGS_INIT:
+		if ('#' == **ppos)
+		    state = PGS_COMMENT;
+		else if (!isspace((unsigned char) **ppos))
+		    state = PGS_DATA;
+		break;
+
+	    default:
+		if ('\n' == **ppos)
+		    state = PGS_INIT;
+	};
+
+	if (PGS_COMMENT == state)
+	    /* skip comment */
+	    continue;
+
+	switch (**ppos) {
+	    case '}':
+		message(MESS_ERROR, "%s:%d unexpected } (missing previous '{')\n", configFile, lineNum);
+		free(globString);
+		return NULL;
+
+	    case '{':
+		/* NUL-terminate globString */
+		assert(globStringPos < globStringAlloc);
+		globString[globStringPos] = '\0';
+		return globString;
+
+	    default:
+		break;
+	}
+
+	/* grow the output buffer if needed */
+	if (globStringPos + 2 > globStringAlloc) {
+	    char *ptr;
+	    globStringAlloc += GLOB_STR_REALLOC_STEP;
+	    ptr = realloc(globString, globStringAlloc);
+	    if (!ptr) {
+		/* out of memory */
+		free(globString);
+		return NULL;
+	    }
+	    globString = ptr;
+	}
+
+	/* copy a single character */
+	globString[globStringPos++] = **ppos;
+    }
+
+    /* premature end of input */
+    message(MESS_ERROR, "%s:%d missing '{' after log files definition\n", configFile, lineNum);
+    free(globString);
+    return NULL;
+}
+
 static int globerr(const char *pathname, int theerr)
 {
+    (void) pathname;
+
     /* A missing directory is not an error, so return 0 */
     if (theerr == ENOTDIR)
         return 0;
@@ -883,10 +986,6 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					newlog->flags |= LOG_FLAG_COMPRESS;
 				} else if (!strcmp(key, "nocompress")) {
 					newlog->flags &= ~LOG_FLAG_COMPRESS;
-				} else if (!strcmp(key, "compress")) {
-					newlog->flags |= LOG_FLAG_COMPRESS;
-				} else if (!strcmp(key, "nocompress")) {
-					newlog->flags &= ~LOG_FLAG_COMPRESS;
 				} else if (!strcmp(key, "delaycompress")) {
 					newlog->flags |= LOG_FLAG_DELAYCOMPRESS;
 				} else if (!strcmp(key, "nodelaycompress")) {
@@ -988,9 +1087,9 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							!strcmp(key, "maxsize")) {
 					unsigned long long size = 0;
 					char *opt = key;
-							
-					if ((key = isolateValue(configFile, lineNum, opt, &start,
-							&buf, length)) != NULL) {
+
+					key = isolateValue(configFile, lineNum, opt, &start, &buf, length);
+					if (key && key[0]) {
 						int l = strlen(key) - 1;
 						if (key[l] == 'k' || key[l] == 'K') {
 							key[l] = '\0';
@@ -1051,7 +1150,25 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 				} else if (!strcmp(key, "monthly")) {
 					newlog->criterium = ROT_MONTHLY;
 				} else if (!strcmp(key, "weekly")) {
+					unsigned weekday;
+					char tmp;
 					newlog->criterium = ROT_WEEKLY;
+					free(key);
+					key = isolateLine(&start, &buf, length);
+					if (key == NULL || key[0] == '\0') {
+						/* default to Sunday if no argument was given */
+						newlog->weekday = 0;
+						continue;
+					}
+
+					if (1 == sscanf(key, "%u%c", &weekday, &tmp) && weekday <= 7) {
+						/* use the selected weekday, 7 means "once per week" */
+						newlog->weekday = weekday;
+						continue;
+					}
+					message(MESS_ERROR, "%s:%d bad weekly directive '%s'\n",
+							configFile, lineNum, key);
+					goto error;
 				} else if (!strcmp(key, "yearly")) {
 					newlog->criterium = ROT_YEARLY;
 				} else if (!strcmp(key, "rotate")) {
@@ -1083,6 +1200,19 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						}
 					}
 					else continue;
+				} else if (!strcmp(key, "minage")) {
+					free(key);
+					if ((key = isolateValue
+						(configFile, lineNum, "minage count", &start,
+						&buf, length)) != NULL) {
+						newlog->rotateMinAge = strtoul(key, &chptr, 0);
+						if (*chptr || newlog->rotateMinAge < 0) {
+							message(MESS_ERROR, "%s:%d bad minimum age '%s'\n",
+								configFile, lineNum, start);
+							RAISE_ERROR();
+						}
+					}
+					else continue; 
 				} else if (!strcmp(key, "maxage")) {
 					free(key);
 					if ((key = isolateValue
@@ -1156,23 +1286,76 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							while (isspace((unsigned char)*endtag) && *endtag)
 								endtag++;
 						} else {
-							free_2d_array(tabooExts, tabooCount);
+							free_2d_array(tabooPatterns, tabooCount);
 							tabooCount = 0;
-							tabooExts = malloc(1);
+							/* realloc of NULL is safe by definition */
+							tabooPatterns = NULL;
 						}
 
-
 						while (*endtag) {
+							int bytes;
+							char *pattern = NULL;
+
 							chptr = endtag;
 							while (!isspace((unsigned char)*chptr) && *chptr != ',' && *chptr)
 								chptr++;
 
-							tabooExts = realloc(tabooExts, sizeof(*tabooExts) *
+							tabooPatterns = realloc(tabooPatterns, sizeof(*tabooPatterns) *
 										(tabooCount + 1));
-							tabooExts[tabooCount] = malloc(chptr - endtag + 1);
-							strncpy(tabooExts[tabooCount], endtag,
-								chptr - endtag);
-							tabooExts[tabooCount][chptr - endtag] = '\0';
+							bytes = asprintf(&pattern, "*%.*s", (int)(chptr - endtag), endtag);
+
+							/* should test for malloc() failure */
+							assert(bytes != -1);
+							tabooPatterns[tabooCount] = pattern;
+							tabooCount++;
+
+							endtag = chptr;
+							if (*endtag == ',')
+								endtag++;
+							while (*endtag && isspace((unsigned char)*endtag))
+								endtag++;
+						}
+					}
+					else continue;
+				} else if (!strcmp(key, "taboopat")) {
+					if (newlog != defConfig) {
+						message(MESS_ERROR,
+							"%s:%d taboopat may not appear inside "
+							"of log file definition\n", configFile,
+							lineNum);
+						state = STATE_ERROR;
+						continue;
+					}
+					free(key);
+					if ((key = isolateValue(configFile, lineNum, "taboopat", &start,
+							&buf, length)) != NULL) {
+						endtag = key;
+						if (*endtag == '+') {
+							endtag++;
+							while (isspace((unsigned char)*endtag) && *endtag)
+								endtag++;
+						} else {
+							free_2d_array(tabooPatterns, tabooCount);
+							tabooCount = 0;
+							/* realloc of NULL is safe by definition */
+							tabooPatterns = NULL;
+						}
+
+						while (*endtag) {
+							int bytes;
+							char *pattern = NULL;
+
+							chptr = endtag;
+							while (!isspace((unsigned char)*chptr) && *chptr != ',' && *chptr)
+								chptr++;
+
+							tabooPatterns = realloc(tabooPatterns, sizeof(*tabooPatterns) *
+										(tabooCount + 1));
+							bytes = asprintf(&pattern, "%.*s", (int)(chptr - endtag), endtag);
+
+							/* should test for malloc() failure */
+							assert(bytes != -1);
+							tabooPatterns[tabooCount] = pattern;
 							tabooCount++;
 
 							endtag = chptr;
@@ -1189,17 +1372,21 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							&buf, length)) != NULL) {
 
 						message(MESS_DEBUG, "including %s\n", key);
-						if (++recursion_depth > MAX_NESTING) {
+						if (recursion_depth >= MAX_NESTING) {
 							message(MESS_ERROR, "%s:%d include nesting too deep\n",
 									configFile, lineNum);
-							--recursion_depth;
-							goto error;
-						}
-						if (readConfigPath(key, newlog)) {
-							--recursion_depth;
-							goto error;
+							logerror = 1;
+							continue;
 						}
+
+						++recursion_depth;
+						rv = readConfigPath(key, newlog);
 						--recursion_depth;
+
+						if (rv) {
+							logerror = 1;
+							continue;
+						}
 					}
 					else continue;
 				} else if (!strcmp(key, "olddir")) {
@@ -1223,7 +1410,21 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					message(MESS_DEBUG, "extension is now %s\n",
 						newlog->extension);
 
+				} else if (!strcmp(key, "addextension")) {
+					if ((key = isolateValue
+						(configFile, lineNum, "addextension name", &start,
+							&buf, length)) != NULL) {
+						freeLogItem (addextension);
+						newlog->addextension = key;
+						key = NULL;
+					}
+					else continue;
+
+					message(MESS_DEBUG, "addextension is now %s\n",
+						newlog->addextension);
+
 				} else if (!strcmp(key, "compresscmd")) {
+					char *compresscmd_base;
 					freeLogItem (compress_prog);
 
 					if (!
@@ -1242,6 +1443,18 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					message(MESS_DEBUG, "compress_prog is now %s\n",
 						newlog->compress_prog);
 
+					compresscmd_base = strdup(basename(newlog->compress_prog));
+					/* we check whether we changed the compress_cmd. In case we use the apropriate extension
+					   as listed in compress_cmd_list */
+					for(i = 0; i < compress_cmd_list_size; i++) {
+						if (!strcmp(compress_cmd_list[i].cmd, compresscmd_base)) {
+							freeLogItem (compress_ext);
+							newlog->compress_ext = strdup((char *)compress_cmd_list[i].ext);
+							message(MESS_DEBUG, "compress_ext was changed to %s\n", newlog->compress_ext);
+							break;
+						}
+					}
+					free(compresscmd_base);
 				} else if (!strcmp(key, "uncompresscmd")) {
 					freeLogItem (uncompress_prog);
 
@@ -1313,6 +1526,8 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
                                                                            || *start == '~'
 #endif
                                                                            ) {
+				char *glob_string;
+				size_t glob_count;
 				in_config = 0;
 				if (newlog != defConfig) {
 					message(MESS_ERROR, "%s:%d unexpected log filename\n",
@@ -1335,43 +1550,29 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 				if ((newlog = newLogInfo(defConfig)) == NULL)
 					goto error;
 
-				endtag = start;
-				while (endtag - buf < length && *endtag != '{' && *endtag != '}' && *endtag != '\0') {
-					endtag++;}
-				if (endtag - buf > length)
-					continue;
-				if (*endtag == '}') {
-					message(MESS_ERROR, "%s:%d unexpected } (missing previous '{')\n", configFile,
-						lineNum);
-					goto error;
-				}
-				if (*endtag == '{') {
-					in_config = 1;
-				}
-				else {
-					message(MESS_ERROR, "%s:%d missing '{' after log files definition\n", configFile,
-						lineNum);
-					goto error;
-				}
-				char *key = strndup(start, endtag - start);
-				start = endtag;
+				glob_string = parseGlobString(configFile, lineNum, buf, length, &start);
+				if (glob_string)
+				    	in_config = 1;
+				else
+				    	/* error already printed */
+				    	goto error;
 
-				if (poptParseArgvString(key, &argc, &argv)) {
+				if (poptParseArgvString(glob_string, &argc, &argv)) {
 				message(MESS_ERROR, "%s:%d error parsing filename\n",
 					configFile, lineNum);
-				free(key);
+				free(glob_string);
 				goto error;
 				} else if (argc < 1) {
 				message(MESS_ERROR,
 					"%s:%d { expected after log file name(s)\n",
 					configFile, lineNum);
-				free(key);
+				free(glob_string);
 				goto error;
 				}
 
 				newlog->files = NULL;
 				newlog->numFiles = 0;
-				for (argNum = 0; argNum < argc && logerror != 1; argNum++) {
+				for (argNum = 0; argNum < argc; argNum++) {
 				if (globerr_msg) {
 					free(globerr_msg);
 					globerr_msg = NULL;
@@ -1403,9 +1604,9 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 									globResult.
 									gl_pathc));
 
-				for (i = 0; i < globResult.gl_pathc; i++) {
+				for (glob_count = 0; glob_count < globResult.gl_pathc; glob_count++) {
 					/* if we glob directories we can get false matches */
-					if (!lstat(globResult.gl_pathv[i], &sb) &&
+					if (!lstat(globResult.gl_pathv[glob_count], &sb) &&
 					S_ISDIR(sb.st_mode)) {
 						continue;
 					}
@@ -1414,11 +1615,11 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						log = log->list.tqe_next) {
 					for (k = 0; k < log->numFiles; k++) {
 						if (!strcmp(log->files[k],
-							globResult.gl_pathv[i])) {
+							globResult.gl_pathv[glob_count])) {
 						message(MESS_ERROR,
 							"%s:%d duplicate log entry for %s\n",
 							configFile, lineNum,
-							globResult.gl_pathv[i]);
+							globResult.gl_pathv[glob_count]);
 						logerror = 1;
 						goto duperror;
 						}
@@ -1426,14 +1627,14 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					}
 
 					newlog->files[newlog->numFiles] =
-					strdup(globResult.gl_pathv[i]);
+					strdup(globResult.gl_pathv[glob_count]);
 					newlog->numFiles++;
 				}
 		duperror:
 				globfree(&globResult);
 				}
 
-				newlog->pattern = key;
+				newlog->pattern = glob_string;
 
 				free(argv);
 
@@ -1461,15 +1662,17 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 			if (newlog->oldDir) {
 				for (i = 0; i < newlog->numFiles; i++) {
 					char *ld;
-					int rv;
-					dirName = ourDirName(newlog->files[i]);
+					char *dirpath;
+
+					dirpath = strdup(newlog->files[i]);
+					dirName = dirname(dirpath);
 					if (stat(dirName, &sb2)) {
 						if (!(newlog->flags & LOG_FLAG_MISSINGOK)) {
 							message(MESS_ERROR,
 								"%s:%d error verifying log file "
 								"path %s: %s\n", configFile, lineNum,
 								dirName, strerror(errno));
-							free(dirName);
+							free(dirpath);
 							goto error;
 						}
 						else {
@@ -1479,13 +1682,13 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 								"but missingok is set, so this is not an error.\n",
 								configFile, lineNum,
 								dirName, strerror(errno));
-							free(dirName);
+							free(dirpath);
 							continue;
 						}
 					}
 					ld = alloca(strlen(dirName) + strlen(newlog->oldDir) + 2);
 					sprintf(ld, "%s/%s", dirName, newlog->oldDir);
-					free(dirName);
+					free(dirpath);
 
 					if (newlog->oldDir[0] != '/') {
 						dirName = ld;
@@ -1494,11 +1697,22 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						dirName = newlog->oldDir;
 					}
 
-					rv = stat(dirName, &sb);
-					if (rv) {
+					if (stat(dirName, &sb)) {
 						if (errno == ENOENT && newlog->flags & LOG_FLAG_OLDDIRCREATE) {
-							if (mkpath(dirName, newlog->olddirMode,
-								newlog->olddirUid, newlog->olddirGid)) {
+							int ret;
+							if (newlog->flags & LOG_FLAG_SU) {
+								if (switch_user(newlog->suUid, newlog->suGid) != 0) {
+									goto error;
+								}
+							}
+							ret = mkpath(dirName, newlog->olddirMode,
+								newlog->olddirUid, newlog->olddirGid);
+							if (newlog->flags & LOG_FLAG_SU) {
+								if (switch_user_back() != 0) {
+									goto error;
+								}
+							}
+							if (ret) {
 								goto error;
 							}
 						}
@@ -1634,6 +1848,10 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 				key = NULL;
 			}
 			break;
+		default:
+			message(MESS_DEBUG,
+				"%s: %d: readConfigFile() unknown state\n",
+				configFile, lineNum);
 	}
 	if (key) {
 		free(key);
@@ -1654,10 +1872,10 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
 	munmap(buf, (size_t) length);
 	close(fd);
-    return 0;
+    return logerror;
 error:
-	if (key)
-		free(key);
+	/* free is a NULL-safe operation */
+	free(key);
 	munmap(buf, (size_t) length);
 	close(fd);
     return 1;
